//===========================================================================
//
//  Filename  : ganttview.cpp
//  Purpose   : Composite Gantt View for QtGantt. Comprises a QListView, a
//              QSplitter and a GanttChart object. This is the "new" view of the
//              Gantt chart/table.
//  Copyright : Copyright (C) 2000 Matt J. Gumbley <matt@gumbley.demon.co.uk>
//  Original  : by Gunnstein Lye
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//===========================================================================


#include <qlistview.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qevent.h>
#include <qsplitter.h>
#include <qstring.h>
#include <qheader.h>
#include <qscrollview.h>
#include <qlayout.h>

#include "task.h"
#include "project.h"
#include "ganttview.h"
#include "ganttchart.h"
#include "taskinfodialogimpl.h"

GanttView::GanttView ( Project *proj, QWidget *parent, const char *name )
    : QWidget( parent, name )
{
	project = proj;

	// I have given up with splitters for the time being - they're just too
	// awkward. or is it that there's something I'm not understanding (more
	// likely!)
	//
	//splitter = new QSplitter( this );

	// Task table creation
	QFontMetrics fm( fontMetrics() );
	//listview = new QListView( splitter, "gantt table" );
	listview = new QListView( this, "gantt table" );

	listview->addColumn("Task ID");
	listview->setColumnWidthMode(0, QListView::Maximum);
	listview->addColumn("Description");
	listview->setColumnWidthMode(1, QListView::Manual);
	listview->setColumnAlignment(1, AlignLeft);
	listview->addColumn("Duration");
	listview->setColumnWidthMode(2, QListView::Maximum);
	listview->setColumnAlignment(2, AlignRight);
	listview->setSelectionMode( QListView::Extended );
	// The middle column is truncated to fit between the left and right
	// columns
	int two_outsides = listview->columnWidth(0) + listview->columnWidth(2);
	// KLUDGE! the width of the main widget is hard-coded to 790 in
	// main.cpp, and the listview is half of that. I haven't been able to
	// determine the width of the listview at all - listview->width() gives
	// 100, which just ain't right.
	listview->setColumnWidth(1, (790/2) - two_outsides);

	listview->setFocusPolicy( StrongFocus );
	listview->setAllColumnsShowFocus( TRUE );
	listview->setSorting( -1 );

	// Disallow sorting of the table by clicking on headers
	QHeader *qh = listview->header();
	QObject::disconnect(qh, SIGNAL(sectionClicked( int )), 
		listview, SLOT(changeSortColumn( int )) );

	// Selection/deselection of tasks cause toolbar enable/disables
	QObject::connect(listview, SIGNAL(selectionChanged()),
		this, SLOT(slotSelectionChanged()));

	// Double click causes the task information dialog to pop up for that
	// task
	QObject::connect(listview, SIGNAL(doubleClicked(QListViewItem *)),
		this, SLOT(slotTaskInfo(QListViewItem *)));

	//scrollview = new QScrollView( splitter );
	scrollview = new QScrollView( this );
	// The scrollview vertical scrolling will be managed by listening to
	// movements in the listview's vertical scroll bar. Hopefully!
	scrollview->setHScrollBarMode( QScrollView::Auto );
	scrollview->setVScrollBarMode( QScrollView::AlwaysOff );
	// Qt Experts: how do you make the scrollview take up all the space on
	// the RHS of the splitter?
	ganttchart = new GanttChart( project, 0, "gantt chart" );

	// a little kludge to determine the height of list items...
	listview->clear();
	QListViewItem *lvi = new QListViewItem( listview );
	int taskheight = lvi->height();
	listview->clear();
	ganttchart->setTaskHeight( taskheight );

	// The following should work, but seems to return a height that is
	// greater than the header... so kludge it for now.
	//ganttchart->setHeaderHeight( listview->header()->height() );
	ganttchart->setHeaderHeight( taskheight + 2 );

	ganttchart->setBackgroundColor( "white" );
	scrollview->addChild( ganttchart );

	QHBoxLayout *hLayout = new QHBoxLayout( this );
	hLayout->addWidget(listview);
	hLayout->addWidget(scrollview);
	hLayout->activate();
}


GanttView::~GanttView ()
{
	if (listview)
		delete listview;
	if (ganttchart)
		delete ganttchart;
}

void GanttView::populate()
{
	int  cpt;
	// Qstring for set text in QListViewItem
	QString  task_nb;
	QString  task_de;

	listview->clear();

	//connect(listview, SIGNAL(doubleClicked(QListViewItem *)), this, 
	//		SLOT(handleClick(QListViewItem *)));
	// populating the list
	if ( this->project )
	{
		for (cpt=(project->numtasks ); cpt--; cpt>1 )
		{
			task_nb.sprintf("%d", project->tasks[cpt]->num);
			task_de.sprintf("%d", project->tasks[cpt]->durationlength);
		
			new QListViewItem( listview, task_nb, 
					*project->tasks[cpt]->description,
					task_de );
		}
	}
}

void GanttView::redraw()
{
	listview->setMinimumSize( 400,this->height());
	ganttchart->setMinimumSize( (this->width()-400),this->height() );
	//	listview->repaint();
	this->populate();
}


void GanttView::paint( QPainter *painter )
{
	listview->setMinimumSize( 400,this->height());
	ganttchart->setMinimumSize( (this->width()-400),this->height() );
	//	listview->repaint();
	this->populate();
}

void GanttView::resizeEvent( QResizeEvent * )
{
	this->redraw();
}



void GanttView::changedaypix( int dpix )
{
	ganttchart->changedaypix( dpix );
}

void GanttView::setweekdetail( int newweekdetail )
{
	ganttchart->setweekdetail( newweekdetail );
}


void GanttView::slotSelectionChanged( )
{
	qDebug("GanttView::slotSelectionChanged");
}

void GanttView::slotTaskInfo( QListViewItem * item )
{
	int index = (item->itemPos()/item-> height());
	qDebug("GanttView::slotTaskInfo, Task # %d", index);
	TaskInfoDialogImpl tid( this, "task info" );
	tid.SetView( TaskInfoDialogImpl::GENERAL );
	tid.SetTask(project->tasks[index]);
	tid.exec();
}

