//===========================================================================
//
//  Filename  : mainwidget.cpp
//  Purpose   : Main Application Controller Widget for QtGantt
//  Copyright : Copyright (C) 2000 Matt J. Gumbley <matt@gumbley.demon.co.uk>
//  Original  : by Gunnstein Lye
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//===========================================================================

#include <stream.h>

#include <qapplication.h>
#include <qmainwindow.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qpixmap.h>
#include <qstatusbar.h>
#include <qmenubar.h>
#include <qtabbar.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qwidget.h>
#include <qframe.h>
#include <qtabwidget.h>
#include <qpushbutton.h>
#include <qscrollbar.h>
#include <qspinbox.h>
#include <qaccel.h>
#include <qfile.h>
#include <qprinter.h>
#include <qlayout.h>

// The MainWidget contains these view-widgets, layered on a QWidgetStack.
#include "gantt.h" // the "print preview" Gantt table and chart in one
#include "ganttview.h" // Gantt Table, Splitter, and Gantt Chart
#include "taskview.h" // Task Editor View
#include "calendarview.h" // Calendar View

// The dialogs
#include "taskinfodialogimpl.h"
#include "projectinfodialogimpl.h"

#include "mainwidget.h"
#include "project.h"

#include "pixmaps/newfile.xpm"
#include "pixmaps/openfile.xpm"
#include "pixmaps/savefile.xpm"
#include "pixmaps/printfile.xpm"
#include "pixmaps/linktasks.xpm"
#include "pixmaps/unlinktasks.xpm"
#include "pixmaps/infotask.xpm"
#include "pixmaps/notetask.xpm"
#include "pixmaps/aftertask.xpm"
#include "pixmaps/gototask.xpm"

MainWidget::MainWidget( char *filenamearg, QWidget *parent, const char *name )
        : QMainWindow( parent, name )
	{
	int ctrlH = 25; // might be better the same height as the menu?
	frameT = 5; // thickness between window edge and frame edge

	//
	// Create the Model
	//
	project = new Project( filenamearg );

	//
	// Create the menus
	//
	_menubar = menuBar(); // new QMenuBar( this );
	_menubar->setSeparator( QMenuBar::InWindowsStyle );
	QRect menuRect = _menubar->frameRect();
	frameY = menuRect.height() + ctrlH + (2 * frameT);

	//
	// Populate the menus
	//
	_filemenu = new QPopupMenu;
	CHECK_PTR( _filemenu );
	// All the menu slots are part of the MainWidget, which is the
	// controller. Project is the model; Gantt is one view.
	_filemenu->insertItem( "&New", this, SLOT( slotNewFile() ));
	_filemenu->insertItem( "&Open...", this, SLOT( slotOpenFile() ));
	_filemenu->insertItem( "&Refresh", this, SLOT( slotRefreshFile() ));
	_filemenu->insertItem( "&Save", this, SLOT( slotSaveFile() ));
	_filemenu->insertSeparator();
	_filemenu->insertItem( "Prop&erties", this, SLOT( slotPropertiesFile() ));
	_filemenu->insertSeparator();
	_filemenu->insertItem( "&Print", this, SLOT( slotPrintFile() ));
	_filemenu->insertSeparator();
	_filemenu->insertItem( "&Quit", this, SLOT( slotQuit() ));

	_options = new QPopupMenu;
	CHECK_PTR( _options );
	weekdetailID = _options->insertItem( "&Week detail", this, 
			SLOT( slottoggleweekdetail()) );
	isWeekdetail = FALSE;
	_options->setCheckable( TRUE ); 

	_viewmenu = new QPopupMenu;
	CHECK_PTR( _viewmenu );
	_viewmenu->insertItem( "Zoom in", this, SLOT( slotzoomIn()) );
	_viewmenu->insertItem( "Zoom out", this, SLOT( slotzoomOut()) );
	_viewmenu->insertSeparator();
	_viewmenu->insertItem( "&Options", _options );
	_viewmenu->insertSeparator();
	// these three need making checkable
	previewViewID = _viewmenu->insertItem( "Print Preview", this, 
			SLOT( slotPreviewView()) );
	_viewmenu->setItemChecked( previewViewID, TRUE );
	ganttViewID = _viewmenu->insertItem( "Gantt View", this, 
			SLOT( slotGanttView()) );
	taskViewID = _viewmenu->insertItem( "Task Editor", this, 
			SLOT( slotTaskView()) );
	calViewID = _viewmenu->insertItem( "Calendar View", this, 
			SLOT( slotCalendarView()) );
	_viewmenu->setCheckable( TRUE );

	_helpmenu = new QPopupMenu;
	CHECK_PTR( _helpmenu );
	_helpmenu->insertItem( "&About QtGantt", this, SLOT( slotAbout() ));

	_menubar->insertItem( "&File", _filemenu );
	_menubar->insertItem( "&View", _viewmenu );
	_menubar->insertSeparator( );
	_menubar->insertItem( "&Help", _helpmenu );

	//
	// Create the toolbar
	//
	QToolBar * toolbar = new QToolBar( this, "My Toolbar" );

	QPixmap newfilePixmap = QPixmap( newfile_xpm );
	QToolButton * newtoolbutton = new QToolButton( newfilePixmap, 
		"New Project", 0, this, SLOT(slotNewFile()), 
		toolbar, "Create a new project");

	QPixmap openfilePixmap = QPixmap( openfile_xpm );
	QToolButton * opentoolbutton = new QToolButton( openfilePixmap,
		"Open an existing project", 0, this, SLOT(slotOpenFile()), 
		toolbar, "Open Project");

	QPixmap savefilePixmap = QPixmap( savefile_xpm );
	QToolButton * savetoolbutton = new QToolButton( savefilePixmap,
		"Save the current project", 0, this, SLOT(slotSaveFile()), 
		toolbar, "Save Project");

	QPixmap printfilePixmap = QPixmap( printfile_xpm );
	QToolButton * printtoolbutton = new QToolButton( printfilePixmap,
		"Print the current project", 0, this, SLOT(slotPrintFile()), 
		toolbar, "Print Project");

	toolbar->addSeparator();

	QPixmap linktasksPixmap = QPixmap( linktasks_xpm );
	QToolButton * linktasksbutton = new QToolButton( linktasksPixmap,
		"Link the selected tasks", 0, this, SLOT(slotLinkTasks()), 
		toolbar, "Link Tasks");

	QPixmap unlinktasksPixmap = QPixmap( unlinktasks_xpm );
	QToolButton * unlinktasksbutton = new QToolButton( unlinktasksPixmap,
		"Unlink the selected tasks", 0, this, SLOT(slotUnlinkTasks()), 
		toolbar, "Unlink Tasks");

	toolbar->addSeparator();

	QPixmap infotaskPixmap = QPixmap( infotask_xpm );
	QToolButton * infotaskbutton = new QToolButton( infotaskPixmap,
		"Change Task Details", 0, this, SLOT(slotInfoTask()), 
		toolbar, "Task Details");

	QPixmap notetaskPixmap = QPixmap( notetask_xpm );
	QToolButton * notetaskbutton = new QToolButton( notetaskPixmap,
		"Change Task Note", 0, this, SLOT(slotNoteTask()), 
		toolbar, "Task Note");

	QPixmap aftertaskPixmap = QPixmap( aftertask_xpm );
	QToolButton * aftertaskbutton = new QToolButton( aftertaskPixmap,
		"Change Task Predecessors", 0, this, SLOT(slotAfterTask()), 
		toolbar, "Task Predecessors");

	toolbar->addSeparator();

	QPixmap gototaskPixmap = QPixmap( gototask_xpm );
	QToolButton * gototaskbutton = new QToolButton( gototaskPixmap,
		"Goto selected task", 0, this, SLOT(slotGotoTask()), 
		toolbar, "Goto selected task");

	toolbar->addSeparator();

	// Assign accelerators
	f5acc = new QAccel( this );    // Draws gantt when F5 is pressed
	f5acc->connectItem( f5acc->insertItem(Key_F5), this, 
			SLOT( slotRefreshFile() ) );
	
	plusacc = new QAccel( this );    // Zooms in when + is pressed
	plusacc->connectItem( plusacc->insertItem(Key_Plus), this, 
			SLOT( slotzoomIn() ) );
	minusacc = new QAccel( this );    // Zooms out when - is pressed
	minusacc->connectItem( minusacc->insertItem(Key_Minus), this, 
			SLOT( slotzoomOut() ) );

	// Create scroll bar control
	scrolltextpix = new QScrollBar( 0, 300, 1, 10, 150, 
			QScrollBar::Horizontal, this, "scrolltextpix" );
	int stpW = 150;
	scrolltextpix->setMinimumSize( stpW, ctrlH );
	connect( scrolltextpix, SIGNAL( valueChanged(int) ), this, 
			SLOT(slotchangetextpix(int)) );

	// Create zoom control
	spindaypix = new QSpinBox( MINDAYPIX, MAXDAYPIX, 1, this, "spindaypix");
	spindaypix->setMinimumSize( 50, ctrlH );
	spindaypix->setValue( 10 );
	connect( spindaypix, SIGNAL( valueChanged(int) ), this, 
			SLOT(slotchangedaypix(int)) );

	// Create the tabwidget. This holds all the main widgets, and the view
	// is switched between them using the menu or tab bar.
	tabwidget = new QTabWidget( this, "Tab Bar" );
	tabwidget->setGeometry(2,2,10,25);
	tabwidget->setMargin(1);
	tabwidget->setTabPosition( QTabWidget::Bottom );
	connect( tabwidget, SIGNAL( selected(const QString &) ), this, 
			SLOT(slotViewTabSelected(const QString &)) );

	//
	// Create the statusbar
	//
	_statusbar = statusBar();
	_statusbar->message( (const char *) "Welcome to QtGantt!", 10000);

	// Create all views
	// The Gantt object is the "old" view, useful for print preview
	gantt = new Gantt( project, tabwidget, "gantt" );
	tabwidget->addTab( gantt, "Print Preview" );

	// The GanttView is the "new" view, comprising QListView of tasks,
	// QSplitter, and GanttChart.
	ganttview = new GanttView( project, tabwidget, "ganttview" );
	tabwidget->addTab( ganttview, "Gantt View" );
	tabwidget->showPage( ganttview );

	// The Task Editor allows editing of tasks
	taskview = new TaskView( project, tabwidget, "taskview" );
	tabwidget->addTab( taskview, "Task Editor" );

	// The CalendarView is another view, by time.
	calendarview = new CalendarView( project, tabwidget, "calendar" );
	tabwidget->addTab( calendarview, "Calendar" );

	setCentralWidget( tabwidget );

	//
	// Layout management
	//
	// Create a vertical layout to contain all widgets and sub-layouts.

	QVBoxLayout * vLayout = new QVBoxLayout( this, 2 );
	vLayout->setMenuBar( _menubar );

	// Create a horizontal layout to hold the toolbar, spin control and
	// zoom scroll control
	// FIXME: there's something not quite right here - this code works well
	// on QT2.1, but not on QT2.2. Could the QT experts help me please?!
	QHBoxLayout * hTopLayout = new QHBoxLayout( vLayout, 15, "horizontal" );

	hTopLayout->addWidget( toolbar, 0 );
	hTopLayout->addSpacing( 20 );
	hTopLayout->addStretch( 3 );

	hTopLayout->addWidget( scrolltextpix, 0 );
	hTopLayout->addSpacing( 20 );

	hTopLayout->addWidget( spindaypix, 0 );

	// Fill the rest of the horizontal box up with no stretchable space, so
	// that the widgets get their minimum width and are pushed to the left.
	hTopLayout->addStretch( 0 );

	// Finished with the top layout

	// The tabwidget grabs all the space in the middle. Set vertical
	// stretch to 10 to let the tabwidget stretch vertically. It will
	// stretch horizontally because there are no widgets beside it to take
	// up the horizontal stretch.
	vLayout->addWidget( tabwidget, 10 );

	vLayout->addWidget( _statusbar );

	// Activate that layout!
	vLayout->activate();
	}


void MainWidget::slottoggleweekdetail()
{
	isWeekdetail = !isWeekdetail;
	_options->setItemChecked( weekdetailID, isWeekdetail );
	gantt->setweekdetail(isWeekdetail);
	ganttview->setweekdetail(isWeekdetail);
}

void MainWidget::slotzoomIn()
{
	int s = spindaypix->value();
	if (s < MAXDAYPIX ) {
		spindaypix->setValue( ++s );
		gantt->changedaypix(s);
		ganttview->changedaypix(s);
	}
}

void MainWidget::slotzoomOut()
{
	int s = spindaypix->value();
	if (s > MINDAYPIX ) {
		spindaypix->setValue( --s );
		gantt->changedaypix(s);
		ganttview->changedaypix(s);
	}
}

void MainWidget::resizeEvent( QResizeEvent * )
	{
	}


void MainWidget::slotNewFile()
{
	int t;
	if ( project->isChanged() )
	{
		switch (QMessageBox::information( this, "QtGantt",
				"This file contains unsaved work\n"
				"Do you want to save it before creating a new file?\n",
				"&Save", "&Don't save", "&Cancel", 0, 2))
		{
		case 0: // Save
			slotSaveFile();
			break;
		case 1: // Don't Save
			break;
		case 2: // Cancel
			return;
		}
	}

	// Reset and repaint
	project->reset();
	gantt->repaint(); // may have to be refresh
	taskview->repaint();
}

void MainWidget::slotOpenFile()
{
	if ( project->isChanged() )
	{
		switch (QMessageBox::information( this, "QtGantt",
				"This file contains unsaved work\n"
				"Do you want to save it before opening another file?\n",
				"&Save", "&Don't save", "&Cancel", 0, 2))
		{
		case 0: // Save
			slotSaveFile();
			break;
		case 1: // Don't Save
			break;
		case 2: // Cancel
			return;
		}
	}

	// Open file dialog; Open file (which resets before reading)
	QString filename( QFileDialog::getOpenFileName( 0, "*.gantt" ) );
	if ( !filename.isNull() )             // if cancel was not pressed
	{
		project->openFilename( filename );
		gantt->repaint(); // may have to be refresh
		taskview->redraw();
	}
}

void MainWidget::slotSaveFile()
{
	qDebug("slotSaveFile");
	if ( ! project->isChanged() )
	{
		QMessageBox::information( this, "QtGantt",
			"The project has not changed: no save necessary.\n");
		return;
	}

	if ( ! project->HasFilename() )
	{
		QString filename( QFileDialog::getSaveFileName( 0, "*.gantt" ) );
		if ( !filename.isNull() )	// if cancel was not pressed
		{
			if (project->SetFilename( filename ))
			{
				qDebug("SetFilename returned false");
				return;
			}
		}
	}
	else
	{
		qDebug("Calling save...");
		if (!project->save())
		{
			QMessageBox::information( this, "QtGantt",
			"The save failed.\n");
		}
		else
		{
			_statusbar->message("Saved OK", 10000);
		}

		qDebug("Finished calling save...");
	}
}

void MainWidget::slotRefreshFile()
{
	if ( project->isChanged() )
	{
		switch (QMessageBox::information( this, "QtGantt",
				"This file has changed.\n"
				"Refreshing will abandon your changes.\n"
				"Do you still want to refresh?\n",
				"&Yes", "&No", 0, 1))
		{
		case 0: // Yes, reload
			break;
		case 1: // Don't reload
			return;
		}
	}
	project->refresh();
	gantt->repaint(); // may have to be refresh
	taskview->redraw();
	}

void MainWidget::slotPropertiesFile()
{
	ProjectInfoDialogImpl pid( this, "project info dialog", project );
	pid.exec();
}

void MainWidget::slotPrintFile()
{
	QPrinter * printer = new QPrinter();
	gantt->print( printer );
}


void MainWidget::slotPreviewView()
{
	// We can be invoked by the user changing the current tab, or by the
	// menu. If the user changes the tab, then the view is automatically
	// changed - we only need to change the view if we're currently showing
	// a different one.
	if (tabwidget->currentPage() != gantt )
		tabwidget->showPage( gantt );
	_viewmenu->setItemChecked( previewViewID, TRUE );
	_viewmenu->setItemChecked( ganttViewID, FALSE );
	_viewmenu->setItemChecked( taskViewID, FALSE );
	_viewmenu->setItemChecked( calViewID, FALSE );
	const char * msg = "If there are more tasks than can fit, they don't display correctly.";
	qWarning(msg);
	_statusbar->message(msg, 10000);
}

void MainWidget::slotGanttView()
{
	// We can be invoked by the user changing the current tab, or by the
	// menu. If the user changes the tab, then the view is automatically
	// changed - we only need to change the view if we're currently showing
	// a different one.
	if (tabwidget->currentPage() != ganttview )
		tabwidget->showPage( ganttview );
	_viewmenu->setItemChecked( previewViewID, FALSE );
	_viewmenu->setItemChecked( ganttViewID, TRUE );
	_viewmenu->setItemChecked( taskViewID, FALSE );
	_viewmenu->setItemChecked( calViewID, FALSE );
	const char * msg = "There are slight dependency arrow misdrawings in this view";
	qWarning(msg);
	_statusbar->message(msg, 10000);
}

void MainWidget::slotTaskView()
{
	// We can be invoked by the user changing the current tab, or by the
	// menu. If the user changes the tab, then the view is automatically
	// changed - we only need to change the view if we're currently showing
	// a different one.
	if (tabwidget->currentPage() != taskview )
		tabwidget->showPage( taskview );
	_viewmenu->setItemChecked( previewViewID, FALSE );
	_viewmenu->setItemChecked( ganttViewID, FALSE );
	_viewmenu->setItemChecked( taskViewID, TRUE );
	_viewmenu->setItemChecked( calViewID, FALSE );
	const char * msg = "Don't forget to Update Tasks when editing";
	qWarning(msg);
	_statusbar->message(msg, 10000);
}

void MainWidget::slotCalendarView()
{
	// We can be invoked by the user changing the current tab, or by the
	// menu. If the user changes the tab, then the view is automatically
	// changed - we only need to change the view if we're currently showing
	// a different one.
	if (tabwidget->currentPage() != calendarview )
		tabwidget->showPage( calendarview );
	_viewmenu->setItemChecked( previewViewID, FALSE );
	_viewmenu->setItemChecked( ganttViewID, FALSE );
	_viewmenu->setItemChecked( taskViewID, FALSE );
	_viewmenu->setItemChecked( calViewID, TRUE );
	const char * msg = "Calendar view is unfinished";
	qWarning(msg);
	_statusbar->message(msg, 10000);
}

// Invoked when the user selects a tab - we emit a signal to the appropriate
// handler above, to update the status message (and do any other tasks which may
// be needed when the tab changes)
void MainWidget::slotViewTabSelected(const QString &tablabel)
{
	QWidget * w = tabwidget->currentPage();
	if (w == gantt)
		slotPreviewView(); // FIX: should be emit
	else if (w == ganttview)
		slotGanttView(); // FIX: should be emit
	else if (w == taskview)
		slotTaskView(); // FIX: should be emit
	else if (w == calendarview)
		slotCalendarView(); // FIX: should be emit
	else 
		qWarning("Odd value in slotViewTabSelected");
}

void MainWidget::slotLinkTasks()
{
	qDebug("Creating the tid");
	TaskInfoDialogImpl tid( this, "link tasks dialog" );
	tid.exec();
}

void MainWidget::slotUnlinkTasks()
{
}

void MainWidget::slotInfoTask()
{
}

void MainWidget::slotNoteTask()
{
}

void MainWidget::slotAfterTask()
{
}

void MainWidget::slotGotoTask()
{
}

void MainWidget::slotchangetextpix(int tpix)
{
	gantt->changetextpix(tpix);
}

void MainWidget::slotchangedaypix(int dpix)
{
	gantt->changedaypix(dpix);
	ganttview->changedaypix(dpix);
}



void MainWidget::slotQuit()
{
	if ( project->isChanged() )
	{
		switch (QMessageBox::information( this, "QtGantt",
				"This file contains unsaved work\n"
				"Do you want to save it before quitting?\n",
				"&Save", "&Don't save", "&Cancel", 0, 2))
		{
		case 0: // Save
			slotSaveFile();
			break;
		case 1: // Don't Save
			break;
		case 2: // Cancel
			return;
		}
	}
	// emit to qApp? perhaps change this when I understand signals/slots
	// better
	qApp->quit();
}

void MainWidget::slotAbout()
{
	QMessageBox::information( this, "About QtGantt",
				"QtGantt - a project management system\n"
				"Copyright 2000 by Gunnstein Lye, Matt Gumbley\n"
				"and the gantt-devel@geekstuff.co.uk\n"
				"mailing list\n"
				);
}

